import math
import random
import unittest
from sklearn.utils.estimator_checks import *
import pandas as pd
import category_encoders as encoders
from category_encoders import WOEEncoder

# Subroutines
def create_array(n_rows=1000, extras=False, has_none=True):
    """
    Creates a numpy dataset with some categorical variables
    :return:
    """

    ds = [[
        random.random(),
        random.random(),
        random.choice(['A', 'B', 'C']),
        random.choice(['A', 'B', 'C', 'D']) if extras else random.choice(['A', 'B', 'C']),
        random.choice(['A', 'B', 'C', None, np.nan]) if has_none else random.choice(['A', 'B', 'C']),
        random.choice(['A'])
    ] for _ in range(n_rows)]

    return np.array(ds)

def create_dataset(n_rows=1000, extras=False, has_none=True):
    """
    Creates a dataset with some categorical variables
    """

    ds = [[
        random.random(),                                                                        # Floats
        random.choice([float('nan'), float('inf'), float('-inf'), -0, 0, 1, -1, math.pi]),      # Floats with edge scenarios
        row,                                                                                    # Unique integers
        str(row),                                                                               # Unique strings
        random.choice(['A']),                                                                   # Invariant
        random.choice(['A', 'B_b', 'C_c_c']),                                                   # Strings with underscores to test reverse_dummies()
        random.choice(['A', 'B', 'C', None]) if has_none else random.choice(['A', 'B', 'C']),   # None
        random.choice(['A', 'B', 'C', 'D']) if extras else random.choice(['A', 'B', 'C']),      # With a new string value
        random.choice([12, 43, -32])                                                            # Numbers everywhere
    ] for row in range(n_rows)]

    df = pd.DataFrame(ds, columns=['float', 'float_edge', 'unique_int', 'unique_str', 'invariant', 'underscore', 'none', 'extra', 321])
    return df

def verify_numeric(X_test):
    for dt in X_test.dtypes:
        numeric = False
        if np.issubdtype(dt, np.dtype(int)) or np.issubdtype(dt, np.dtype(float)):
            numeric = True
        assert numeric

def verify_inverse_transform(x, x_inv):
    """
    Verify x is equal to x_inv. The test returns true for NaN.equals(NaN) as it should.

    """
    assert x.equals(x_inv)

# Data definitions
np_X = create_array(n_rows=100)
np_X_t = create_array(n_rows=50, extras=True)
np_y = np.random.randn(np_X.shape[0]) > 0.5
np_y_t = np.random.randn(np_X_t.shape[0]) > 0.5
df_X = create_dataset(n_rows=100)
df_X_t = create_dataset(n_rows=50, extras=True)
df_y = pd.DataFrame(np_y)
df_y_t = pd.DataFrame(np_y_t)


class NumbersTest(unittest.TestCase):

    def test_np(self):
        for encoder_name in encoders.__all__:
            with self.subTest(encoder_name=encoder_name):

                # Encode a numpy array
                enc = getattr(encoders, encoder_name)()
                enc.fit(np_X, np_y)
                verify_numeric(enc.transform(np_X_t))


    def test_classification(self):
        for encoder_name in encoders.__all__:
            with self.subTest(encoder_name=encoder_name):
                cols = ['unique_str', 'underscore', 'extra', 'none', 'invariant', 321]

                enc = getattr(encoders, encoder_name)(cols=cols)
                enc.fit(df_X, np_y)
                verify_numeric(enc.transform(df_X_t))

                enc = getattr(encoders, encoder_name)(verbose=1)
                enc.fit(df_X, np_y)
                verify_numeric(enc.transform(df_X_t))

                enc = getattr(encoders, encoder_name)(drop_invariant=True)
                enc.fit(df_X, np_y)
                verify_numeric(enc.transform(df_X_t))

                enc = getattr(encoders, encoder_name)(return_df=False)
                enc.fit(df_X, np_y)
                self.assertTrue(isinstance(enc.transform(df_X_t), np.ndarray))
                self.assertEqual(enc.transform(df_X_t).shape[0], df_X_t.shape[0], 'Row count must not change')

    def test_impact_encoders(self):
        for encoder_name in ['LeaveOneOutEncoder', 'TargetEncoder', 'WOEEncoder']:
            with self.subTest(encoder_name=encoder_name):

                # Encode a numpy array and transform with the help of the target
                enc = getattr(encoders, encoder_name)()
                enc.fit(np_X, np_y)
                verify_numeric(enc.transform(np_X_t, np_y_t))

                # Target is a DataFrame
                enc = getattr(encoders, encoder_name)()
                enc.fit(df_X, df_y)
                verify_numeric(enc.transform(df_X_t, df_y_t))

    # Only hashing encoder fails...
    def test_error_handling(self):
        for encoder_name in encoders.__all__:
            with self.subTest(encoder_name=encoder_name):

                # We exclude some columns
                X = create_dataset(n_rows=100)
                X = X.drop(['unique_str', 'none'], axis=1)
                X_t = create_dataset(n_rows=50, extras=True)
                X_t = X_t.drop(['unique_str', 'none'], axis=1)

                # New value
                enc = getattr(encoders, encoder_name)(verbose=1, return_df=True, handle_unknown='error')
                enc.fit(X, np_y)
                with self.assertRaises(ValueError):
                    _ = enc.transform(X_t)

                # Illegal state, we have to first train the encoder...
                enc = getattr(encoders, encoder_name)()
                with self.assertRaises(ValueError):
                    enc.transform(X)

                # Wrong count of attributes
                enc = getattr(encoders, encoder_name)()
                enc.fit(X, np_y)
                with self.assertRaises(ValueError):
                    enc.transform(X_t.iloc[:, 0:3])

                # No cols
                enc = getattr(encoders, encoder_name)(cols=[])
                enc.fit(X, np_y)
                self.assertTrue(enc.transform(X_t).equals(X_t))

    def test_sklearn_compliance(self):
        for encoder_name in encoders.__all__:
            with self.subTest(encoder_name=encoder_name):

                check_transformer_general(encoder_name, getattr(encoders, encoder_name)())
                check_transformers_unfitted(encoder_name, getattr(encoders, encoder_name)())

    # Only BaseN does not raise the exception...
    def test_inverse_transform(self):
        # We do not allow None in these data
        X = create_dataset(n_rows=100, has_none=False)
        X_t = create_dataset(n_rows=50, has_none=False)
        X_t_extra = create_dataset(n_rows=50, extras=True, has_none=False)
        cols = ['underscore', 'none', 'extra', 321]

        for encoder_name in ['BaseNEncoder', 'BinaryEncoder', 'OneHotEncoder', 'OrdinalEncoder']:
            with self.subTest(encoder_name=encoder_name):
                enc = getattr(encoders, encoder_name)(verbose=1, cols=cols)
                enc.fit(X, np_y)
                verify_inverse_transform(X_t, enc.inverse_transform(enc.transform(X_t)))
                with self.assertRaises(ValueError):
                    _ = enc.inverse_transform(enc.transform(X_t_extra))

                enc = getattr(encoders, encoder_name)(verbose=1, cols=cols, drop_invariant=True, handle_unknown='error')
                enc.fit(X)
                verify_inverse_transform(X_t, enc.inverse_transform(enc.transform(X_t)))

    # Encoder specific tests
    def test_woe(self):
        cols = ['unique_str', 'underscore', 'extra', 'none', 'invariant', 321]

        # Balanced label with balanced features
        X_balanced = pd.DataFrame(data=['1', '1', '1', '2', '2', '2'], columns=['col1'])
        y_balanced = [True, False, True, False, True, False]
        enc = encoders.WOEEncoder()
        enc.fit(X_balanced, y_balanced)
        X1 = enc.transform(X_balanced)
        self.assertTrue(all(X1.sum() < 0.001), "When the class label is balanced, WoE should sum to 0 in each transformed column")

        enc = WOEEncoder(cols=cols)
        enc.fit(df_X, np_y)
        X1 = enc.transform(df_X_t)
        verify_numeric(X1[cols])
        self.assertTrue(np.isfinite(X1[cols].values).all(), 'There must not be any NaN, inf or -inf in the transformed columns')
        self.assertEqual(len(list(df_X_t)), len(list(X1)), 'The count of attributes must not change')
        self.assertEqual(len(df_X_t), len(X1), 'The count of rows must not change')
        X2 = enc.transform(df_X_t, np_y_t)
        verify_numeric(X2)
        self.assertTrue(np.isfinite(X2[cols].values).all(), 'There must not be any NaN, inf or -inf in the transformed columns')
        self.assertEqual(len(list(df_X_t)), len(list(X2)), 'The count of attributes must not change')
        self.assertEqual(len(df_X_t), len(X2), 'The count of rows must not change')
        X3 = enc.transform(df_X, np_y)
        verify_numeric(X3)
        self.assertTrue(np.isfinite(X3[cols].values).all(), 'There must not be any NaN, inf or -inf in the transformed columns')
        self.assertEqual(len(list(df_X)), len(list(X3)), 'The count of attributes must not change')
        self.assertEqual(len(df_X), len(X3), 'The count of rows must not change')
        self.assertTrue(X3['unique_str'].var() < 0.001, 'The unique string column must not be predictive of the label')
        X4 = enc.fit_transform(df_X, np_y)
        verify_numeric(X4)
        self.assertTrue(np.isfinite(X4[cols].values).all(), 'There must not be any NaN, inf or -inf in the transformed columns')
        self.assertEqual(len(list(df_X)), len(list(X4)), 'The count of attributes must not change')
        self.assertEqual(len(df_X), len(X4), 'The count of rows must not change')
        self.assertTrue(X4['unique_str'].var() < 0.001, 'The unique string column must not be predictive of the label')

        enc = WOEEncoder()
        enc.fit(df_X, np_y)
        X1 = enc.transform(df_X_t)
        self.assertEqual(len(list(df_X_t)), len(list(X1)), 'The count of attributes must not change')
        self.assertEqual(len(df_X_t), len(X1), 'The count of rows must not change')
        verify_numeric(X1)
        X2 = enc.transform(df_X_t, np_y_t)
        verify_numeric(X2)
        self.assertEqual(len(list(df_X_t)), len(list(X2)), 'The count of attributes must not change')
        self.assertEqual(len(df_X_t), len(X2), 'The count of rows must not change')

        # Seed
        enc = WOEEncoder(cols=cols, random_state=2001, randomized=True)
        enc.fit(df_X, np_y)
        X1 = enc.transform(df_X_t, np_y_t)
        X2 = enc.transform(df_X_t, np_y_t)
        self.assertTrue(X1.equals(X2), "When the seed is given, the results must be identical")
        verify_numeric(X1)
        verify_numeric(X2)

        # Invariant target
        y_invariant = [True, True, True, True, True, True]
        enc = encoders.WOEEncoder()
        with self.assertRaises(ValueError):
            enc.fit(X_balanced, y_invariant)

        # Branch coverage unit tests - no cols
        enc = WOEEncoder(cols=[])
        enc.fit(df_X, np_y)
        self.assertTrue(enc.transform(df_X_t).equals(df_X_t))

        # Missing values in the target
        y_missing = [True, True, None, True, True, True]
        enc = WOEEncoder()
        with self.assertRaises(ValueError):
            enc.fit(X_balanced, y_missing)

        # Impute missing
        enc = WOEEncoder(impute_missing=False)
        enc.fit(df_X, np_y)
        X1 = enc.transform(df_X_t)
        verify_numeric(X1)
        self.assertTrue(X1.isnull().values.any())
        self.assertEqual(len(list(df_X_t)), len(list(X1)), 'The count of attributes must not change')
        self.assertEqual(len(df_X_t), len(X1), 'The count of rows must not change')

        X2 = enc.transform(df_X_t, np_y_t)
        verify_numeric(X2)
        self.assertTrue(X1.isnull().values.any())
        self.assertEqual(len(list(df_X_t)), len(list(X2)), 'The count of attributes must not change')
        self.assertEqual(len(df_X_t), len(X2), 'The count of rows must not change')

    def test_loo(self):
        df = pd.DataFrame({
            'color': ["a", "a", "a", "b", "b", "b"],
            'outcome': [1, 0, 0, 1, 0, 1]})

        X = df.drop('outcome', axis=1)
        y = df.drop('color', axis=1)

        ce_leave = encoders.LeaveOneOutEncoder(cols=['color'], randomized=False)
        obtained = ce_leave.fit_transform(X, y['outcome'])

        self.assertEquals([0.0, 0.5, 0.5, 0.5, 1.0, 0.5], list(obtained['color']))
